/* oh_volume.c - OpenHome Volume service routines.
 *
 * Copyright (C) 2005-2007   Ivo Clarysse
 * Copyright (C) 2013 Andrey Demenev
 *
 * This file is part of GMediaRender.
 *
 * GMediaRender is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GMediaRender is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GMediaRender; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "oh_volume.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include <glib.h>

#include "upnp.h"
#include "upnp_device.h"
#include "output.h"
#include "oh_time.h"
#include "volume.h"

#define VOLUME_TYPE "urn:av-openhome-org:service:Volume:1"
#define VOLUME_SERVICE_ID "urn:av-openhome:serviceId:Volume"

#define VOLUME_SCPD_URL "/upnp/openhomevolumeSCPD.xml"
#define VOLUME_CONTROL_URL "/upnp/control/openhomevolume1"
#define VOLUME_EVENT_URL "/upnp/event/openhomevolume1"

typedef enum {
	VOLUME_VAR_VOLUME,
	VOLUME_VAR_BALANCE,
	VOLUME_VAR_FADE,
	VOLUME_VAR_MUTE,
	VOLUME_VAR_VOLUME_LIMIT,
	VOLUME_VAR_VOLUME_MAX,
	VOLUME_VAR_VOLUME_UNITY,
	VOLUME_VAR_VOLUME_STEPS,
	VOLUME_VAR_VOLUME_MILLI_DB_PER_STEP,
	VOLUME_VAR_BALANCE_MAX,
	VOLUME_VAR_FADE_MAX,
	
	VOLUME_VAR_UNKNOWN,
	VOLUME_VAR_COUNT
} volume_variable_t;

enum {
	VOLUME_CMD_VOLUME,
	VOLUME_CMD_SET_VOLUME,
	VOLUME_CMD_VOLUME_INC,
	VOLUME_CMD_VOLUME_DEC,
	VOLUME_CMD_VOLUME_LIMIT,
	VOLUME_CMD_BALANCE,
	VOLUME_CMD_SET_BALANCE,
	VOLUME_CMD_BALANCE_INC,
	VOLUME_CMD_BALANCE_DEC,
	VOLUME_CMD_FADE,
	VOLUME_CMD_SET_FADE,
	VOLUME_CMD_FADE_INC,
	VOLUME_CMD_FADE_DEC,
	VOLUME_CMD_MUTE,
	VOLUME_CMD_SET_MUTE,
	VOLUME_CMD_CHARACTERISTICS,

	VOLUME_CMD_UNKNOWN,
	VOLUME_CMD_COUNT
};


static const char *volume_default_values[] = {
	[VOLUME_VAR_VOLUME] = VOLUME_MAX,
	[VOLUME_VAR_BALANCE] = "0",
	[VOLUME_VAR_FADE] = "0",
	[VOLUME_VAR_MUTE] = "0",
	[VOLUME_VAR_VOLUME_LIMIT] = VOLUME_MAX,
	[VOLUME_VAR_VOLUME_MAX] = VOLUME_MAX,
	[VOLUME_VAR_VOLUME_UNITY] = VOLUME_MAX,
	[VOLUME_VAR_VOLUME_STEPS] = VOLUME_MAX,
	[VOLUME_VAR_VOLUME_MILLI_DB_PER_STEP] = "0",
	[VOLUME_VAR_BALANCE_MAX] = "0",
	[VOLUME_VAR_FADE_MAX] = "0",
	
	[VOLUME_VAR_UNKNOWN] = NULL,
};


static struct var_meta volume_var_meta[] = {
	[VOLUME_VAR_VOLUME] =                   { "Volume", SENDEVENT_YES, DATATYPE_UI4, NULL, NULL },
	[VOLUME_VAR_BALANCE] =                  { "Balance", SENDEVENT_YES, DATATYPE_I4, NULL, NULL },
	[VOLUME_VAR_FADE] =                     { "Fade", SENDEVENT_YES, DATATYPE_I4, NULL, NULL },
	[VOLUME_VAR_MUTE] =                     { "Mute", SENDEVENT_YES, DATATYPE_BOOLEAN, NULL, NULL },
	[VOLUME_VAR_VOLUME_LIMIT] =             { "VolumeLimit", SENDEVENT_YES, DATATYPE_UI4, NULL, NULL },
	[VOLUME_VAR_VOLUME_MAX] =               { "VolumeMax", SENDEVENT_YES, DATATYPE_UI4, NULL, NULL },
	[VOLUME_VAR_VOLUME_UNITY] =             { "VolumeUnity", SENDEVENT_YES, DATATYPE_UI4, NULL, NULL },
	[VOLUME_VAR_VOLUME_STEPS] =             { "VolumeSteps", SENDEVENT_YES, DATATYPE_UI4, NULL, NULL },
	[VOLUME_VAR_VOLUME_MILLI_DB_PER_STEP] = { "VolumeMilliDbPerStep", SENDEVENT_YES, DATATYPE_UI4, NULL, NULL },
	[VOLUME_VAR_BALANCE_MAX] =              { "BalanceMax", SENDEVENT_YES, DATATYPE_UI4, NULL, NULL },
	[VOLUME_VAR_FADE_MAX] =                 { "FadeMax", SENDEVENT_YES, DATATYPE_UI4, NULL, NULL },
	[VOLUME_VAR_UNKNOWN] =                  { NULL, SENDEVENT_NO, DATATYPE_UNKNOWN, NULL, NULL },
};

static struct argument *arguments_char[] = {
        & (struct argument) { "VolumeMax", PARAM_DIR_OUT, VOLUME_VAR_VOLUME_MAX },
        & (struct argument) { "VolumeUnity", PARAM_DIR_OUT, VOLUME_VAR_VOLUME_UNITY },
        & (struct argument) { "VolumeSteps", PARAM_DIR_OUT, VOLUME_VAR_VOLUME_STEPS },
        & (struct argument) { "VolumeMilliDbPerStep", PARAM_DIR_OUT, VOLUME_VAR_VOLUME_MILLI_DB_PER_STEP },
        & (struct argument) { "BalanceMax", PARAM_DIR_OUT, VOLUME_VAR_BALANCE_MAX },
        & (struct argument) { "FadeMax", PARAM_DIR_OUT, VOLUME_VAR_FADE_MAX },
        NULL
};

static struct argument *arguments_volume[] = {
        & (struct argument) { "Value", PARAM_DIR_OUT, VOLUME_VAR_VOLUME },
        NULL
};

static struct argument *arguments_set_volume[] = {
        & (struct argument) { "Value", PARAM_DIR_IN, VOLUME_VAR_VOLUME },
        NULL
};

static struct argument *arguments_balance[] = {
        & (struct argument) { "Value", PARAM_DIR_OUT, VOLUME_VAR_BALANCE },
        NULL
};

static struct argument *arguments_set_balance[] = {
        & (struct argument) { "Value", PARAM_DIR_IN, VOLUME_VAR_BALANCE },
        NULL
};

static struct argument *arguments_fade[] = {
        & (struct argument) { "Value", PARAM_DIR_OUT, VOLUME_VAR_FADE },
        NULL
};

static struct argument *arguments_set_fade[] = {
        & (struct argument) { "Value", PARAM_DIR_IN, VOLUME_VAR_FADE },
        NULL
};

static struct argument *arguments_mute[] = {
        & (struct argument) { "Value", PARAM_DIR_OUT, VOLUME_VAR_MUTE },
        NULL
};

static struct argument *arguments_set_mute[] = {
        & (struct argument) { "Value", PARAM_DIR_IN, VOLUME_VAR_MUTE },
        NULL
};

static struct argument *arguments_volume_limit[] = {
        & (struct argument) { "Value", PARAM_DIR_OUT, VOLUME_VAR_VOLUME_LIMIT },
        NULL
};


static struct argument **argument_list[] = {
	[VOLUME_CMD_VOLUME] = arguments_volume,
	[VOLUME_CMD_SET_VOLUME] = arguments_set_volume,
	[VOLUME_CMD_VOLUME_INC] = NULL,
	[VOLUME_CMD_VOLUME_DEC] = NULL,
	[VOLUME_CMD_VOLUME_LIMIT] = arguments_volume_limit,
	[VOLUME_CMD_BALANCE] = arguments_balance,
	[VOLUME_CMD_SET_BALANCE] = arguments_set_balance,
	[VOLUME_CMD_BALANCE_INC] = NULL,
	[VOLUME_CMD_BALANCE_DEC] = NULL,
	[VOLUME_CMD_FADE] = arguments_fade,
	[VOLUME_CMD_SET_FADE] = arguments_set_fade,
	[VOLUME_CMD_FADE_INC] = NULL,
	[VOLUME_CMD_FADE_DEC] = NULL,
	[VOLUME_CMD_MUTE] = arguments_mute,
	[VOLUME_CMD_SET_MUTE] = arguments_set_mute,
	[VOLUME_CMD_CHARACTERISTICS] = arguments_char,
	
	[VOLUME_CMD_UNKNOWN] = NULL
};


extern struct service volume_service_;
static variable_container_t *state_variables_ = NULL;
static uint32_t current_volume = VOLUME_MAX_VAL;


static void service_lock(void)
{
	if (volume_service_.var_change_collector) {
		UPnPVarChangeCollector_start(volume_service_.var_change_collector);
	}
}

static void service_unlock(void)
{
	if (volume_service_.var_change_collector) {
		UPnPVarChangeCollector_finish(volume_service_.var_change_collector);
	}
}

static int replace_var(volume_variable_t varnum, const char *new_value) {
	return VariableContainer_change(state_variables_, varnum, new_value);
}

static int replace_var_uint(volume_variable_t varnum, unsigned int new_value)
{
	char buf[32];
	sprintf(buf, "%u", new_value);
	return replace_var(varnum, buf);
}

static int no_action(struct action_event *event)
{
	return 0;
}

static int volume(struct action_event *event)
{
	upnp_append_variable(event, VOLUME_VAR_VOLUME, "Value");
	return 0;
}

static int volume_limit(struct action_event *event)
{
	upnp_append_variable(event, VOLUME_VAR_VOLUME_LIMIT, "Value");
	return 0;
}

static int mute(struct action_event *event)
{
	upnp_append_variable(event, VOLUME_VAR_MUTE, "Value");
	return 0;
}

static int zero_value(struct action_event *event)
{
	upnp_add_response(event, "Value", "0");
	return 0;
}

static int set_mute(struct action_event *event)
{
	char *value = upnp_get_string(event, "Value");
	if (value == NULL)
		return -1;
	service_lock();
	replace_var(VOLUME_VAR_MUTE, value);
	if(strcmp("1", value) == 0){
		int volume = 0;
		replace_var_uint(VOLUME_VAR_VOLUME, volume);
		const float decibel = volume_level_to_decibel(volume);
		const double fraction = exp(decibel / 20 * log(10));
		output_set_volume(fraction);
		current_volume = volume;
	}
	service_unlock();
	return 0;
}

static int characteristics(struct action_event *event)
{
	upnp_add_response(event, "VolumeMax", VOLUME_MAX);
	upnp_add_response(event, "VolumeUnity", VOLUME_MAX);
	upnp_add_response(event, "VolumeSteps", VOLUME_MAX);
	upnp_add_response(event, "VolumeMilliDbPerStep", "0");
	upnp_add_response(event, "BalanceMax", "0");
	upnp_add_response(event, "FadeMax", "0");
	return 0;
}

static int update_volume(int volume)
{
	replace_var_uint(VOLUME_VAR_VOLUME, volume);
	const float decibel = volume_level_to_decibel(volume);
	const double fraction = exp(decibel / 20 * log(10));
	output_set_volume(fraction);
	int do_mute = (volume == volume_range.min);
	replace_var(VOLUME_VAR_MUTE, do_mute ? "1" : "0");
	output_set_mute(do_mute);
	current_volume = volume;
	return 0;
}

static int volume_inc(struct action_event *event)
{
	service_lock();
	if (current_volume < volume_range.max) {
		update_volume(current_volume+1);
	}
	service_unlock();
	return 0;
}

static int volume_dec(struct action_event *event)
{
	service_lock();
	if (current_volume > volume_range.min) {
		update_volume(current_volume-1);
	}
	service_unlock();
	return 0;
}

static int set_volume(struct action_event *event)
{
	char *str = upnp_get_string(event, "Value");
	if (str == NULL)
		return -1;

	int vol = current_volume;
	sscanf(str, "%u", &vol);
	free(str);
	if (vol < volume_range.min || vol > volume_range.max) {
		upnp_set_error(event, 800, "Invalid volume");
		return -1;
	}
	service_lock();
	update_volume(vol);
	service_unlock();
	return 0;
}

static struct action volume_actions[] = {
	[VOLUME_CMD_VOLUME] = { "Volume", volume },
	[VOLUME_CMD_SET_VOLUME] = { "SetVolume", set_volume },
	[VOLUME_CMD_VOLUME_INC] = { "VolumeInc", volume_inc },
	[VOLUME_CMD_VOLUME_DEC] = { "VolumeDec", volume_dec },
	[VOLUME_CMD_VOLUME_LIMIT] = { "VolumeLimit", volume_limit },
	[VOLUME_CMD_BALANCE] = { "Balance", zero_value },
	[VOLUME_CMD_SET_BALANCE] = { "SetBalance", no_action },
	[VOLUME_CMD_BALANCE_INC] = { "BalanceInc", no_action },
	[VOLUME_CMD_BALANCE_DEC] = { "BalanceDec", no_action },
	[VOLUME_CMD_FADE] = { "Fade", zero_value },
	[VOLUME_CMD_SET_FADE] = { "SetFade", no_action },
	[VOLUME_CMD_FADE_INC] = { "FadeInc", no_action },
	[VOLUME_CMD_FADE_DEC] = { "FadeDec", no_action },
	[VOLUME_CMD_MUTE] = { "Mute", mute },
	[VOLUME_CMD_SET_MUTE] = { "SetMute", set_mute },
	[VOLUME_CMD_CHARACTERISTICS] = { "Characteristics", characteristics },
	[VOLUME_CMD_UNKNOWN] = {NULL, NULL}
};

struct service *oh_volume_get_service(void) {
#ifdef NO_VOLUME
	return NULL;
#else
	if (volume_service_.variable_container == NULL) {
		state_variables_ =
			VariableContainer_new(VOLUME_VAR_COUNT,
						  volume_var_meta,
					      volume_default_values);
		volume_service_.variable_container = state_variables_;
	}
	return &volume_service_;
#endif
}

void oh_volume_init(struct upnp_device *device) {
#ifndef NO_VOLUME
	// Set initial volume.
	float volume_fraction = 0;
	if (output_get_volume(&volume_fraction) == 0) {
		Log_info("oh volume", "Output inital volume is %f; setting "
			 "control variables accordingly.", volume_fraction);
		float decibel = 20 * log(volume_fraction) / log(10);
		int vol = volume_decibel_to_level(decibel);
		Log_info("oh volume", "Setting volume-db to %.2fdb == #%d",decibel, vol);
		update_volume(vol);
	}
	
	assert(volume_service_.var_change_collector == NULL);
	volume_service_.var_change_collector =
		UPnPVarChangeCollector_new(state_variables_, 
			"",  /* const char *event_xml_namespace - not used, since we do not provide LastChange variable */
			device,
			VOLUME_SERVICE_ID);
#endif
}

struct service volume_service_ = {
	.service_id =           VOLUME_SERVICE_ID,
	.service_type =         VOLUME_TYPE,
	.scpd_url =				VOLUME_SCPD_URL,
	.control_url =			VOLUME_CONTROL_URL,
	.event_url =			VOLUME_EVENT_URL,
	.actions =              volume_actions,
	.action_arguments =     argument_list,
	.variable_container =   NULL,
	.var_change_collector =          NULL,
	.variable_meta =        volume_var_meta,
	.variable_count =       VOLUME_VAR_UNKNOWN,
	.command_count =        VOLUME_CMD_UNKNOWN,
};

