/* upnp_device.c - Generic UPnP device handling
 *
 * Copyright (C) 2005-2007   Ivo Clarysse
 * Copyright (C) 2014        Buffalo
 *
 * This file is part of GMediaRender.
 *
 * GMediaRender is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GMediaRender is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GMediaRender; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>
#include <glib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <libgupnp/gupnp-error.h>

#include "logging.h"

#include "xmlescape.h"
#include "xmldoc.h"
#include "upnp.h"
#include "upnp_device.h"
#include "variable-container.h"

GUPnPContext *context;

// Enable logging of action requests.
//#define ENABLE_ACTION_LOGGING

int upnp_add_response(struct action_event *event,
		      const char *key, const char *value)
{
	assert(event != NULL);
	assert(key != NULL);
	assert(value != NULL);

	if (event->status) {
		return -1;
	}

	gupnp_service_action_set(event->request,key,G_TYPE_STRING,value,NULL);

	/*if (rc != UPNP_E_SUCCESS) {
		/// report custom error
		event->request->ActionResult = NULL;
		event->request->ErrCode = UPNP_SOAP_E_ACTION_FAILED;
		strcpy(event->request->ErrStr, UpnpGetErrorMessage(rc));
		return -1;
	}*/
	return 0;
}

void upnp_append_variable(struct action_event *event,
                          int varnum, const char *paramname)
{
	const char *value;
	struct service *service = event->service;

	assert(event != NULL);
	assert(paramname != NULL);

	VariableContainer_lock(service->variable_container);

	value = VariableContainer_get(service->variable_container, varnum, NULL, NULL);
	assert(value != NULL);   // triggers on invalid variable.
	upnp_add_response(event, paramname, value);

	VariableContainer_unlock(service->variable_container);
}

void upnp_set_error(struct action_event *event, int error_code,
		    const char *format, ...)
{
	/*event->status = -1;

	va_list ap;
	va_start(ap, format);
	event->request->ActionResult = NULL;
	event->request->ErrCode = UPNP_SOAP_E_ACTION_FAILED;
	vsnprintf(event->request->ErrStr, sizeof(event->request->ErrStr),
		  format, ap);

	va_end(ap);*/
	Log_error("upnp", "%s: %s\n", __FUNCTION__, event->request);
}

char *upnp_get_string(struct action_event *event, const char *key)
{
	char *value;
	gupnp_service_action_get(event->request,key,G_TYPE_STRING,&value,NULL);
	return strdup(value != NULL ? value : "");
}

struct service *find_service(struct upnp_device_descriptor *device_def,
                             const char *service_id)
{
	struct service *event_service;
	int serviceNum = 0;

	assert(device_def != NULL);
	assert(service_id != NULL);
	while (event_service =
	       device_def->services[serviceNum], event_service != NULL) {
		if (strcmp(event_service->service_id, service_id) == 0)
			return event_service;
		serviceNum++;
	}
	Log_error("upnp","cannot find service");
	return NULL;
}


int upnp_device_notify(struct upnp_device *device,
                       const char *serviceID,
                       const char **varnames,
                       const char **varvalues, int varcount)
{
	int i;
	int c;
	struct service *srv = find_service(device->upnp_device_descriptor,serviceID);
	if(srv){
		GUPnPServiceInfo *service = srv->info;
		if(varcount <= 0){
		}else if(varcount == 1){
			char *key   = varnames[0];
			char *value = varvalues[0];
			if(key != NULL && value != NULL){
				gupnp_service_notify(GUPNP_SERVICE(service),key,G_TYPE_STRING,value,NULL);
			}
		}else{
			c = 0;
			for(i = 0 ; i < varcount; ++i){
				char *key   = varnames[i];
				char *value = varvalues[i];
				if(key != NULL && value != NULL){
					c++;
				}else{
					Log_error("upnp", "null notify %s %s",key,value);
				}
			}
			if(0 < c){
				gupnp_service_freeze_notify (GUPNP_SERVICE(service));
				for(i = 0 ; i < varcount && i < 16 ; ++i){
					//最大数を制限している。根本的解決策模索中
					char *key   = varnames[i];
					char *value = varvalues[i];
					if(key != NULL && value != NULL){
						gupnp_service_notify(GUPNP_SERVICE(service),key,G_TYPE_STRING,value,NULL);
					}
				}
				gupnp_service_thaw_notify (GUPNP_SERVICE(service));
			}else{
				Log_error("upnp", "nothing to notify");
			}
		}
		return 0;
	}
	return -1;
}


static int handle_var_request(GUPnPService *service, gchar *variable, GValue *gvalue, gpointer user_data){
	struct upnp_device *priv = user_data;
	struct service *srv = find_service(priv->upnp_device_descriptor,
			gupnp_service_info_get_id(service));
	if (srv == NULL) {
		//var_event->ErrCode = UPNP_SOAP_E_INVALID_ARGS;
		return -1;
	}

	VariableContainer_lock(srv->variable_container);

	char *result = NULL;
	const int var_count =
		VariableContainer_get_num_vars(srv->variable_container);
	for (int i = 0; i < var_count; ++i) {
		const char *name;
		const char *value =
			VariableContainer_get(srv->variable_container, i, &name, NULL);
		if (value && strcmp(variable, name) == 0) {
			result = strdup(value);
			break;
		}
	}

	VariableContainer_unlock(srv->variable_container);

	if(result != NULL){
		g_value_init (gvalue, G_TYPE_STRING);
		g_value_set_string (gvalue, result);
		free(result);
	}

//todoエラーと送信
	/*var_event->CurrentVal = result;
	var_event->ErrCode = (result == NULL)
		? UPNP_SOAP_E_INVALID_VAR
		: UPNP_E_SUCCESS;
	Log_info("upnp", "Variable request %s -> %s (%s)",
		 var_event->StateVarName, result, gupnp_service_info_get_id(var_event));*/

	return 0;
}

static int handle_action_request(GUPnPService *service, GUPnPServiceAction *action, gpointer user_data)
{
	struct upnp_device *priv = user_data;
	struct service *event_service;
	struct action *event_action;

	event_service = find_service(priv->upnp_device_descriptor,
			gupnp_service_info_get_id(service));
	event_action = find_action(event_service, gupnp_service_action_get_name(action));

	if (event_action == NULL) {
		char error_str[256];
		sprintf(error_str, "Unknown action '%s' for service '%s'",gupnp_service_action_get_name(action), gupnp_service_info_get_id(service));
		Log_error("upnp", error_str);
		gupnp_service_action_return_error(action,GUPNP_CONTROL_ERROR_INVALID_ACTION,error_str);
		return -1;
	}

	// We want to send the LastChange event only after the action is
	// finished - just to be conservative, we don't know how clients
	// react to get LastChange notifictions while in the middle of
	// issuing an action.
	//
	// So we nest the change collector level here, so that we only send the
	// LastChange after the action is finished ().
	//
	// Note, this is, in fact, only a preparation and not yet working as
	// described above: we are still in the middle
	// of executing the event-callback while sending the last change
	// event implicitly when calling UPnPLastChangeCollector_finish() below.
	// It would be good to enqueue the upnp_device_notify() after
	// the action event is finished.
	if (event_service->var_change_collector) {
		UPnPVarChangeCollector_start(event_service->var_change_collector);
	}

#ifdef ENABLE_ACTION_LOGGING
	{
		char *action_request_xml = NULL;
		if (action->ActionRequest) {
			action_request_xml = ixmlDocumenttoString(
					   action->ActionRequest);
		}
		Log_info("upnp", "Action '%s'; Request: %s",
			 gupnp_service_action_get_name(action), action_request_xml);
		free(action_request_xml);
	}
#endif

	if (event_action->callback) {
		struct action_event event;
		int rc;
		event.request = action;
		event.status = 0;
		event.service = event_service;
                event.device = priv;

		rc = (event_action->callback) (&event);
		if (rc == 0) {
#ifdef ENABLE_ACTION_LOGGING
			if (action->ActionResult) {
				char *action_result_xml = NULL;
				action_result_xml = ixmlDocumenttoString(
						action->ActionResult);
				Log_info("upnp", "Action '%s' OK; Response %s",
					 gupnp_service_action_get_name(action),
					 action_result_xml);
				free(action_result_xml);
			} else {
				Log_info("upnp", "Action '%s' OK",
					 gupnp_service_action_get_name(action));
			}
#endif
			gupnp_service_action_return (action);
		}else{
			gupnp_service_action_return_error(action,GUPNP_CONTROL_ERROR_ACTION_FAILED,"The action failed.");
		}
	} else {
		Log_error("upnp",
			  "Got a valid action, but no handler defined (!)\n"
			  "  ActionName: '%s'\n"
			  "  ServiceID:  '%s'\n",
			  gupnp_service_action_get_name(action),
			  gupnp_service_info_get_id(service));
		gupnp_service_action_return (action);
	}

	if (event_service->var_change_collector) {   // See comment above.
		UPnPVarChangeCollector_finish(event_service->var_change_collector);
	}
	return 0;
}

static gboolean initialize_device(struct upnp_device_descriptor *device_def,
				  struct upnp_device *result_device,
				  const char *network_iface,
				  unsigned short port)
{
	char *doc;
	GUPnPRootDevice *dev;
	GError *error = NULL;

	/* Create the UPnP context */
	context = gupnp_context_new (NULL, network_iface, port, &error);
	if (error) {
		Log_error("upnp", "Error creating the GUPnP context (ip=%s, port=%d): %s (%d)",
			gupnp_context_get_host_ip(context), gupnp_context_get_port(context), error->message, error->code);
		g_error_free (error);
		return FALSE;
	}
	Log_info("upnp", "Registered IP=%s port=%d\n",
			gupnp_context_get_host_ip(context), gupnp_context_get_port(context));

	doc = upnp_create_device_desc(device_def,gupnp_context_get_host_ip(context),gupnp_context_get_port(context));
	/*rc = UpnpRegisterRootDevice2(UPNPREG_BUF_DESC,
				     buf, strlen(buf), 1,
				     &event_handler, result_device,
				     &(result_device->device_handle));*/

	/* Create root device */
	dev = gupnp_root_device_new (context, doc , "/tmp");

	result_device->device_handle = dev;

	return TRUE;
}

struct upnp_device *upnp_device_init(struct upnp_device_descriptor *device_def,
				     const char *network_iface,
				     unsigned short port)
{
	int rc;
	char *buf;
	struct service *srv;
	struct icon *icon_entry;

	assert(device_def != NULL);

	if (device_def->init_function) {
		rc = device_def->init_function();
		if (rc != 0) {
			return NULL;
		}
	}

	struct upnp_device *result_device = malloc(sizeof(*result_device));
	result_device->upnp_device_descriptor = device_def;
	pthread_mutex_init(&(result_device->device_mutex), NULL);

	/* register icons in web server */
	for (int i = 0; (icon_entry = device_def->icons[i]); i++) {
		//webserver_register_file(icon_entry->url, "image/png");
	}

	if (!initialize_device(device_def, result_device, network_iface, port)) {
		//UpnpFinish();
		free(result_device);
		return NULL;
	}

	/* generate and register service schemas in web server */
	for (int i = 0; (srv = device_def->services[i]); i++) {
		GUPnPServiceInfo *service;
		buf = upnp_get_scpd(srv);
		assert(buf != NULL);
		/* Get the switch service from the root device */
		service = gupnp_device_info_get_service
				(GUPNP_DEVICE_INFO (result_device->device_handle), srv->service_type);
		if (!service) {
			Log_error ("upnp","Cannot get %s service\n",srv->service_id);
			return EXIT_FAILURE;
		}
		srv->info = service;
		for(int j=0; j<srv->command_count; j++) {
			struct action *act = &(srv->actions[j]);
			if (act) {
				char detailed_signal[256];
				sprintf(detailed_signal,"action-invoked::%s",act->action_name);
				g_signal_connect (service, detailed_signal, G_CALLBACK (handle_action_request), result_device);
			}
		}
		for(int j=0; j<srv->variable_count; j++) {
			struct var_meta *meta = &(srv->variable_meta[j]);
			if (meta) {
				char detailed_signal[256];
				sprintf(detailed_signal,"query-variable::%s",meta->name);
				g_signal_connect (service, detailed_signal, G_CALLBACK (handle_var_request), result_device);
			}
		}
	}
	gupnp_root_device_set_available (result_device->device_handle, TRUE);

	return result_device;
}

void upnp_device_shutdown(struct upnp_device *device) {
	struct service *srv;
	struct upnp_device_descriptor *device_def = device->upnp_device_descriptor;

	gupnp_root_device_set_available (device->device_handle, FALSE);
	g_object_unref (context);
}


/// ---- code to generate device descriptor

static struct xmlelement *gen_specversion(struct xmldoc *doc,
                                          int major, int minor)
{
        struct xmlelement *top;

        top=xmlelement_new(doc, "specVersion");

        add_value_element_int(doc, top, "major", major);
        add_value_element_int(doc, top, "minor", minor);

        return top;
}


static struct xmlelement *gen_desc_iconlist(struct xmldoc *doc,
					    struct icon **icons) {
	struct xmlelement *top;
	struct xmlelement *parent;
	struct icon *icon_entry;

	top=xmlelement_new(doc, "iconList");

	for (int i = 0; (icon_entry=icons[i]); i++) {
		parent=xmlelement_new(doc, "icon");
		add_value_element(doc,parent,"mimetype", icon_entry->mimetype);
		add_value_element_int(doc,parent,"width",icon_entry->width);
		add_value_element_int(doc,parent,"height",icon_entry->height);
		add_value_element_int(doc,parent,"depth",icon_entry->depth);
		add_value_element(doc,parent,"url",icon_entry->url);
		xmlelement_add_element(doc, top, parent);
	}

	return top;
}


static struct xmlelement *
gen_desc_servicelist(struct upnp_device_descriptor *device_def,
		     struct xmldoc *doc)
{
	int i;
	struct service *srv;
	struct xmlelement *top;
	struct xmlelement *parent;

	top=xmlelement_new(doc, "serviceList");

        for (i=0; (srv = device_def->services[i]); i++) {
		parent = xmlelement_new(doc, "service");
		add_value_element(doc, parent, "serviceType",srv->service_type);
		add_value_element(doc, parent, "serviceId", srv->service_id);
		add_value_element(doc, parent, "SCPDURL", srv->scpd_url);
		add_value_element(doc, parent, "controlURL", srv->control_url);
		add_value_element(doc, parent, "eventSubURL", srv->event_url);
		xmlelement_add_element(doc, top, parent);
        }

	return top;
}



static char *generate_desc(struct upnp_device_descriptor *device_def,const char *ip_address,const int listen_port)
{
	struct xmldoc *doc;
	struct xmlelement *root;
	struct xmlelement *child;
	struct xmlelement *parent;
	const char *output_file="/tmp/description.xml";
	int ret;

	doc = xmldoc_new();

	root=xmldoc_new_topelement(doc, "root", "urn:schemas-upnp-org:device-1-0");
	child=gen_specversion(doc,1,0);
	xmlelement_add_element(doc, root, child);
	parent=xmlelement_new(doc, "device");
	xmlelement_add_element(doc, root, parent);
	add_value_element(doc,parent,"deviceType", device_def->device_type);
	add_value_element(doc,parent,"presentationURL", device_def->presentation_url);
	add_value_element(doc,parent,"friendlyName", device_def->friendly_name);
	add_value_element(doc,parent,"manufacturer", device_def->manufacturer);
	add_value_element(doc,parent,"manufacturerURL", device_def->manufacturer_url);
	add_value_element(doc,parent,"modelDescription", device_def->model_description);
	add_value_element(doc,parent,"modelName", device_def->model_name);
	add_value_element(doc,parent,"modelNumber", device_def->model_number);
	add_value_element(doc,parent,"modelURL", device_def->model_url);
	add_value_element(doc,parent,"UDN", device_def->udn);
	//add_value_element(doc,parent,"serialNumber", device_def->serial_number);
	//add_value_element(doc,parent,"UPC", device_def->upc);
	if (device_def->icons) {
		child=gen_desc_iconlist(doc,device_def->icons);
		xmlelement_add_element(doc,parent,child);
	}
	child=gen_desc_servicelist(device_def, doc);
	xmlelement_add_element(doc, parent, child);


	char str[256];
	sprintf(str,"http://%s:%d/",ip_address,listen_port);
	add_value_element(doc,root,"URLBase",str);

	/* Write document */
	ret = xmlSaveFormatFile(output_file, doc,1);
	if (ret < 0) {
		fprintf(stderr, "Failed to save %s\n", output_file);
	}
	return output_file;
}

char *upnp_create_device_desc(struct upnp_device_descriptor *device_def,const char *ip_address,const int listen_port) {
	char *doc;
	doc = generate_desc(device_def,ip_address,listen_port);
	return doc;
}
