/* xmldoc.h - XML builder abstraction
 *
 * Copyright (C) 2007   Ivo Clarysse
 * Copyright (C) 2014   Buffalo
 *
 * This file is part of GMediaRender.
 *
 * GMediaRender is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GMediaRender is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GMediaRender; if not, write to the Free Software 
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 *
 */ 

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

#include "xmldoc.h"

// The structs are mere placeholders that we internally cast to the
// objects used in the upnplib.

static xmlDoc *to_idoc(struct xmldoc *x) {
	return (xmlDoc*) x;
}

static xmlNode *to_ielem(struct xmlelement *x) {
	return (xmlNode*) x;
}

struct xmldoc *xmldoc_new(void)
{
	xmlDoc *doc;
	doc = xmlNewDoc("1.0");
	return (struct xmldoc*) doc;
}

void xmldoc_free(struct xmldoc *doc)
{
	assert(doc != NULL);
	xmlFreeDoc(doc);
}

char *xmldoc_tostring(struct xmldoc *doc)
{
	char *result = NULL;
	assert(doc != NULL);
	xmlChar *s;
	int size;
	xmlDocDumpFormatMemory(doc, &s, &size, 0);
	result = strdup(s);
	xmlFree(s);
	return result;
}

struct xmldoc *xmldoc_parsexml(const char *xml_text) {
	int size=strlen(xml_text);
	xmlDoc *doc = xmlParseMemory(xml_text,size);
	return (struct xmldoc*) doc;
}

struct xmlelement *xmldoc_new_topelement(struct xmldoc *doc,
                                         const char *elementName,
                                         const char *xmlns)
{
	assert(doc != NULL);
	assert(elementName != NULL);
	xmlNode *node;
	if (xmlns) {
		node = xmlNewNode(NULL,elementName);
		xmlDocSetRootElement(doc, node);
		xmlNewProp(node, "xmlns", xmlns);
	} else {
		node = xmlNewNode(NULL,elementName);
		xmlDocSetRootElement(doc, node);
	}
	return (struct xmlelement *) node;
}

struct xmlelement *xmlelement_new(struct xmldoc *doc, const char *elementName)
{
	assert(doc != NULL);
	assert(elementName != NULL);
	xmlNode *node;
	node = xmlNewNode(NULL,elementName);
	return (struct xmlelement*) node;
}

static const xmlChar *xmlNodeGetName(xmlNode *cur) {
	assert(cur != NULL);
	return cur->name;
}

static struct xmlelement *find_element(xmlNode *node, const char *key) {
	node = xmlFirstElementChild(node);
	for (/**/; node != NULL; node = xmlNextElementSibling(node)) {
		if (strcmp(xmlNodeGetName(node), key) == 0) {
			return (struct xmlelement*) node;
		}
	}
	return NULL;
}

struct xmlelement *find_element_in_doc(struct xmldoc *doc,
				       const char *key) {

	return find_element(xmlDocGetRootElement(doc), key);
}

struct xmlelement *find_element_in_element(struct xmlelement *element,
					   const char *key) {
	return find_element(to_ielem(element), key);
}

char *get_node_value(struct xmlelement *element) {
	xmlNode *node = (xmlNode*) to_ielem(element);
	//node = xmlFirstElementChild(node);
	const char *node_value = (node != NULL
				  ? xmlNodeGetContent(node)
				  : NULL);
	return strdup(node_value != NULL ? node_value : "");
}

void xmlelement_add_element(struct xmldoc *doc,
			    struct xmlelement *parent,
			    struct xmlelement *child)
{
	assert(doc != NULL);
	assert(parent != NULL);
	assert(child != NULL);
	xmlAddChild((xmlNode *) to_ielem(parent),
                             (xmlNode *) to_ielem(child));
}

void xmlelement_add_text(struct xmldoc *doc,
			 struct xmlelement *parent,
			 const char *text)
{
	assert(doc != NULL);
	assert(parent != NULL);
	assert(text != NULL);
	xmlNode *textNode;
	textNode = xmlNewText(text);
	xmlAddChild(to_ielem(parent), textNode);
}

void xmlelement_set_attribute(struct xmldoc *doc,
			      struct xmlelement *element,
			      const char *name,
			      const char *value)
{
	assert(doc != NULL);
	assert(element != NULL);
	assert(name != NULL);
	assert(value != NULL);
	xmlNewProp(to_ielem(element), name, value);
}

void add_value_element(struct xmldoc *doc,
                       struct xmlelement *parent,
                       const char *tagname, const char *value)
{
        struct xmlelement *top;

        top=xmlelement_new(doc, tagname);
        xmlelement_add_text(doc, top, value);
        xmlelement_add_element(doc, parent, top);
}

struct xmlelement *add_attributevalue_element(struct xmldoc *doc,
					      struct xmlelement *parent,
					      const char *tagname,
					      const char *attribute_name,
					      const char *value)
{
        struct xmlelement *top;

        top = xmlelement_new(doc, tagname);
        xmlelement_set_attribute(doc, top, attribute_name, value);
        xmlelement_add_element(doc, parent, top);
	return top;
}

void add_value_element_int(struct xmldoc *doc,
                           struct xmlelement *parent,
                           const char *tagname, int value)
{
        char *buf;

        asprintf(&buf,"%d",value);
        add_value_element(doc, parent, tagname, buf);
        free(buf);
}
void add_value_element_long(struct xmldoc *doc,
                            struct xmlelement *parent,
                            const char *tagname, long long value)
{
        char *buf;

        asprintf(&buf,"%lld",value);
        add_value_element(doc, parent, tagname, buf);
        free(buf);
}

